import random
import math
import itertools
from collections import OrderedDict
import urllib.request
from typing import Dict, Union, Tuple

from knapsack_problem.ref_func import knapsack_standard_solution as knapsack_func, Item, Knapsack

"""
functions with "_static_" in name - correct proven examples of input output for knapsack problem
So, for debugging funcs.py I used it for checking correct behavior, because
output of "_static_" functions write in docstrings and we can compare strings with construction:

.. code:: python

    from operator import attrgetter
    expected_result = data.__doc__.replace('\n', '').replace(' ', '')
    result = sorted(func(*data()), key=attrgetter('name'))
    result = str(result).replace(' ', '')

For all other tests i use random data with as many arguments as I need
For output I use one of funcs, reference function, on which correct I am convinced
It's on ref_func.py
But i still use it for other test for check other parameters, not only correctness
"""


def pack_up_static_knapsack_1() -> Knapsack:
    """
    [

        Item(name='camera', value=6, weight=1),
        Item(name='food', value=9, weight=2),
        Item(name='water', value=10, weight=3)

    ]
    """
    items = (
        Item('water', 10, 3),
        Item('book', 3, 1),
        Item('food', 9, 2),
        Item('jacket', 5, 2),
        Item('camera', 6, 1)
    )
    weight_limit = 6

    return Knapsack(items, weight_limit)


def pack_up_static_knapsack_2() -> Knapsack:
    """
    [

        Item(name='book', value=1, weight=1),
        Item(name='food', value=2, weight=1),
        Item(name='jacket', value=2, weight=2),
        Item(name='water', value=6, weight=4)

    ]
    """
    items = (
        Item('water', 6, 4),
        Item('book', 1, 1),
        Item('food', 2, 1),
        Item('jacket', 2, 2),
        Item('camera', 4, 12)
    )
    weight_limit = 15

    return Knapsack(items, weight_limit)


def pack_up_static_knapsack_3() -> Knapsack:
    """
    [

        Item(name='apple', value=39, weight=40),
        Item(name='beer', value=52, weight=10),
        Item(name='book', value=30, weight=10),
        Item(name='camera', value=32, weight=30),
        Item(name='t-shirt', value=24, weight=15),
        Item(name='tin', value=68, weight=45),
        Item(name='trousers', value=48, weight=10),
        Item(name='umbrella', value=73, weight=40),
        Item(name='water', value=153, weight=200)

    ]
    """
    items = (
        Item("map", 9, 150),
        Item("compass", 13, 35),
        Item("water", 153, 200),
        Item("sandwich", 50, 160),
        Item("glucose", 15, 60),
        Item("tin", 68, 45),
        Item("banana", 27, 60),
        Item("apple", 39, 40),
        Item("cheese", 23, 30),
        Item("beer", 52, 10),
        Item("suntan cream", 11, 70),
        Item("camera", 32, 30),
        Item("t-shirt", 24, 15),
        Item("trousers", 48, 10),
        Item("umbrella", 73, 40),
        Item("waterproof trousers", 42, 70),
        Item("waterproof overclothes", 43, 75),
        Item("note-case", 22, 80),
        Item("sunglasses", 7, 20),
        Item("towel", 18, 12),
        Item("socks", 4, 50),
        Item("book", 30, 10)
    )
    weight_limit = 400

    return Knapsack(items, weight_limit)


def create_dynamic_knapsacks(*, start: int, end: int, step: int = 1) -> \
        Dict[str, Dict[str, Union[Knapsack, Tuple[Item]]]]:
    """
    Dynamic create collections of Items
    as result -> {

        knapsack_*weight_limit*: {

            'input':  Tuple[Item] # generated items
            'output': Knapsack # Knapsack answer by knapsack_standard_solution (for checking with other solutions)

    }
    """

    knapsacks = OrderedDict()
    for i in range(start, end + 1, step):
        weight_limit = i
        knapsack_key = 'knapsack_{weight_limit}'.format(weight_limit=weight_limit)
        knapsacks.setdefault(knapsack_key, dict())

        knapsack_question = _create_knapsack(weight_limit)
        knapsack_answer = tuple(knapsack_func(*knapsack_question))

        knapsacks[knapsack_key]['input'] = knapsack_question
        knapsacks[knapsack_key]['output'] = knapsack_answer

    return knapsacks


def _create_knapsack(weight_limit: int) -> Knapsack:
    words = _get_random_words(limit=weight_limit)

    # create list of unique tuples for smallest spread for items value and weight, depends on weight_limit
    max_spread = math.ceil(math.sqrt(weight_limit) * 2)
    values_items_list = list(itertools.product(range(1, max_spread), repeat=2))
    random.shuffle(values_items_list)

    items = []
    for word in words:
        value, weight = values_items_list.pop()
        item = Item(word, value, weight)
        items.append(item)

    return Knapsack(tuple(items), weight_limit)


def _get_random_words(limit=None):
    word_url = "http://svnweb.freebsd.org/csrg/share/dict/words?view=co&content-type=text/plain"
    response = urllib.request.urlopen(word_url)
    long_txt = response.read().decode()
    words = long_txt.splitlines()[:limit]
    random.shuffle(words)
    while words:
        yield words.pop()



"""
Examples of python realization of solution for knapsack problem
"""
from itertools import combinations
from typing import Tuple, List, Generator, Any, Union, Iterator

from knapsack_problem.ref_func import knapsack_standard_solution, Item


# @profile
def knapsack_1_standard_solution(items: Tuple[Item], weight_limit: int) -> Item:
    """
    https://codereview.stackexchange.com/a/20581

    Solve the knapsack problem by finding the most valuable subsequence
    of items that weighs no more than weight_limit.

    items must be a sequence of pairs (value, weight), where value is a
    number and weight is a non-negative integer.

    weight_limit is a non-negative integer.

    Return a pair whose first element is the sum of values in the most
    valuable subsequence, and whose second element is the subsequence.
    """
    return knapsack_standard_solution(items, weight_limit)


# @profile
def knapsack_2_solution(items: Tuple[Item], weight_limit: int) -> Item:
    """
    my own function, written thanks to the site:
    https://foxford.ru/wiki/informatika/algoritm-ukladki-ryukzaka
    """
    w = weight_limit
    f = [[0] * (w + 1) for i in range(len(items) + 1)]
    for i in range(len(items)):
        weight = items[i].weight
        value = items[i].value
        for j in range(1, w + 1):
            if j >= weight:
                f[i][j] = max(f[i - 1][j], f[i - 1][j - weight] + value)
            else:
                f[i][j] = f[i - 1][j]

    for k in reversed(range(len(items))):
        if f[k][w] != f[k - 1][w]:
            yield items[k]
            w -= items[k].weight


# @profile
def knapsack_3_solution(items: Tuple[Item], weight_limit: int) -> Item:
    """
    Given a list of items with name, value and weight.
    Return the optimal value with total weight <= allowed weight and
    list of picked items.
    https://codereview.stackexchange.com/a/125386
    """

    # find which item are picked
    def fetch_items(k: List[List[int]], weight_limit: int, items: Tuple[Item]):
        for item, weights_p, weights_n in zip(items[::-1], k[-2::-1], k[::-1]):
            if weights_n[weight_limit] != weights_p[weight_limit]:
                yield item
                weight_limit -= item.weight

    k = [
        [0] * (weight_limit + 1)
        for x in range(len(items) + 1)
    ]
    for next_idx, (item, weights) in enumerate(zip(items, k), 1):
        for w, current_weight in enumerate(weights[1:], 1):
            if item.weight <= w:
                k[next_idx][w] = max(
                    item.value + weights[w - item.weight],
                    current_weight
                )
            else:
                k[next_idx][w] = current_weight

    return fetch_items(k, weight_limit, items)


# @profile
def knapsack_4_bruteforce_solution(items: Tuple[Item], weight_limit: int) -> Union[Item, List[None]]:
    """
    Brute force algorithm
    http://rosettacode.org/mw/index.php?title=Knapsack_problem/0-1&action=edit&section=62
    """

    def any_comb(items: Tuple[Item]) -> Generator:
        """return combinations of any length from the items"""
        return (comb
                for r in range(1, len(items) + 1)
                for comb in combinations(items, r)
                )

    def total_value(comb: Tuple[Any]) -> Tuple[int, int]:
        """Totalise a particular combination of items"""
        totwt = totval = 0
        for item, val, wt in comb:
            totwt += wt
            totval += val
        return (totval, -totwt) if totwt <= weight_limit else (0, 0)

    bagged = max(any_comb(items), key=total_value)  # max val or min wt if values equal
    if bagged[0].weight <= weight_limit:
        return bagged
    return []


# @profile
def knapsack_5_dynamic_solution(items: Tuple[Item], weight_limit: int) -> Item:
    """
    Dynamic programming solution
    http://rosettacode.org/mw/index.php?title=Knapsack_problem/0-1&action=edit&section=63
    """

    table = [[0] * (weight_limit + 1) for j in range(len(items) + 1)]

    for j in range(1, len(items) + 1):
        item, val, wt = items[j - 1]
        for w in range(1, weight_limit + 1):
            if wt > w:
                table[j][w] = table[j - 1][w]
            else:
                table[j][w] = max(table[j - 1][w],
                                  table[j - 1][w - wt] + val)

    w = weight_limit
    for j in range(len(items), 0, -1):
        was_added = table[j][w] != table[j - 1][w]
        if was_added:
            item, val, wt = items[j - 1]
            yield items[j - 1]
            w -= wt


# @profile
def knapsack_6_recursive_dynamic_solution(items: Tuple[Item], weight_limit: int) -> Tuple[Item]:
    """
    Recursive dynamic programming algorithm
    http://rosettacode.org/mw/index.php?title=Knapsack_problem/0-1&action=edit&section=64
    """

    def total_value(items: Tuple[Item], weight_limit: int) -> int:
        return sum([x.value for x in items]) if sum([x.weight for x in items]) <= weight_limit else 0

    cache = {}

    def solve(items: Tuple[Item], weight_limit: int) -> Tuple:
        if not items:
            return ()
        if (items, weight_limit) not in cache:
            head = items[0]
            tail = items[1:]
            include = (head,) + solve(tail, weight_limit - head[2])
            dont_include = solve(tail, weight_limit)
            if total_value(include, weight_limit) > total_value(dont_include, weight_limit):
                answer = include
            else:
                answer = dont_include
            cache[(items, weight_limit)] = answer
        return cache[(items, weight_limit)]

    return solve(items, weight_limit)


# @profile
def knapsack_greedy_solution(items: Tuple[Item], weight_limit: int) -> Iterator:
    """
    Return a list of items with the maximum value, subject to the
    constraint that their combined weight must not exceed max_weight.
    Implements the well-known "greedy approximation algorithm" for the knapsack problem
    (first described by George Dantzig in 1957).
    https://codereview.stackexchange.com/a/62871
    """

    def efficiency(item: Item) -> float:
        """Return efficiency of item (its value per unit weight)."""
        return float(item.value) / float(item.weight)

    def pack(item: Item) -> bool:
        # Attempt to pack item; return True if successful.
        if item.weight <= pack.max_weight:
            pack.max_weight -= item.weight
            return True
        else:
            return False

    pack.max_weight = weight_limit
    return filter(pack, sorted(items, key=efficiency, reverse=True))


if __name__ == '__main__':
    import time

    from data import pack_up_static_knapsack_3

    funcs = [
        knapsack_1_standard_solution,
        knapsack_2_solution,
        knapsack_3_solution,
        knapsack_4_bruteforce_solution,
        knapsack_5_dynamic_solution,
        knapsack_6_recursive_dynamic_solution,
        knapsack_greedy_solution
    ]

    for func in funcs:
        list(func(*pack_up_static_knapsack_3()))
        time.sleep(0.3)



import unittest
from typing import List
from numbers import Integral

from binary_tree.funcs import BisectNodeClass, SingleNodeClass, TwoNodeClass, BaseNodeClass
from binary_tree.data import datalist_100, datalist_1000, datalist_10000


class TestNodes(unittest.TestCase):
    data = [datalist_100, datalist_1000, datalist_10000]

    def tree_testing(self, instance: BaseNodeClass, datalist: List[Integral]):
        for i in datalist:
            instance.add_node(i)
        result = list(instance.tree_data())
        self.assertEqual(result, sorted(datalist))

    def test_data_SingleNodeClass(self):
        for datalist in self.data:
            node = SingleNodeClass()
            self.tree_testing(node, datalist)

    def test_data_TwoNodeClass(self):
        for datalist in self.data:
            node = TwoNodeClass()
            self.tree_testing(node, datalist)

    def test_data_BisectNodeClass(self):
        for datalist in self.data:
            node = BisectNodeClass()
            self.tree_testing(node, datalist)


if __name__ == '__main__':
    unittest.main()



import unittest
from operator import attrgetter
from typing import Callable
import time

from knapsack_problem.data import (
    pack_up_static_knapsack_1,
    pack_up_static_knapsack_2,
    pack_up_static_knapsack_3,
    create_dynamic_knapsacks
)
from knapsack_problem.funcs import (
    knapsack_1_standard_solution,
    knapsack_2_solution,
    knapsack_3_solution,
    knapsack_4_bruteforce_solution,
    knapsack_5_dynamic_solution,
    knapsack_6_recursive_dynamic_solution,
    knapsack_greedy_solution
)
from knapsack_problem.ref_func import knapsack_standard_solution

TOO_LONG = 60 * 0.1  # time in seconds for break loop of testing functions


class TestKnapsackProblem1StaticData(unittest.TestCase):
    dataset = [pack_up_static_knapsack_1, pack_up_static_knapsack_2, pack_up_static_knapsack_3]

    @classmethod
    def setUpClass(cls) -> None:
        print('*' * 15, cls.__name__, '*' * 15)

    def main_part(self, *, func: Callable) -> None:
        print('#' * 5, func.__name__, '#' * 5)
        for data in self.dataset:
            expected_result = data.__doc__.replace('\n', '').replace(' ', '')

            start = time.monotonic()
            result = sorted(func(*data()), key=attrgetter('name'))
            result = str(result).replace(' ', '')
            end = time.monotonic() - start
            self.assertEqual(expected_result, result)
            print('Knapsack with {} items was completed in {:2f} seconds'.format(len(data().items), end))

    def test_knapsack_1_solution(self) -> None:
        self.main_part(func=knapsack_1_standard_solution)

    def test_knapsack_2_solution(self) -> None:
        self.main_part(func=knapsack_2_solution)

    def test_knapsack_3_solution(self) -> None:
        self.main_part(func=knapsack_3_solution)

    def test_knapsack_4_solution(self) -> None:
        self.main_part(func=knapsack_4_bruteforce_solution)

    def test_knapsack_5_solution(self) -> None:
        self.main_part(func=knapsack_5_dynamic_solution)

    def test_knapsack_6_solution(self) -> None:
        self.main_part(func=knapsack_6_recursive_dynamic_solution)

    def test_knapsack_greedy_solution(self) -> None:
        print('#' * 5, knapsack_greedy_solution.__name__, '#' * 5)
        for data in self.dataset:
            knapsack = data()

            start = time.monotonic()
            raw_result = list(knapsack_greedy_solution(*knapsack))
            self.assertTrue(knapsack.weight_limit >= sum(i.weight for i in raw_result))
            end = time.monotonic() - start

            raw_expected_result = knapsack_standard_solution(*knapsack)
            expected_result = sum(i.value for i in raw_expected_result) * 0.9
            result = sum(i.value for i in raw_result)

            self.assertTrue(result >= expected_result)
            print('Knapsack with {} items was completed in {:2f} seconds'.format(len(data().items), end))


class TestKnapsackProblem2DynamicData(unittest.TestCase):
    dataset = create_dynamic_knapsacks(start=10, end=50, step=4)

    @classmethod
    def setUpClass(cls) -> None:
        print('*' * 15, cls.__name__, '*' * 15)

    def main_part(self, func: Callable) -> None:
        print('#' * 5, func.__name__, '#' * 5)
        for data in self.dataset.values():
            expected_result = sum(i.value for i in data['output'])

            start = time.monotonic()
            raw_result = func(*data['input'])
            result = sum(i.value for i in raw_result)
            end = time.monotonic() - start

            self.assertTrue(data['input'].weight_limit >= sum(i.weight for i in raw_result))
            self.assertEqual(result, expected_result)

            print('Knapsack with {} items was completed in {:2f} seconds'.format(len(data['input'].items), end))
            if end > TOO_LONG:
                print('Function {func_name} working too long'.format(func_name=func.__name__))
                break

    def test_knapsack_1_solution(self) -> None:
        self.main_part(func=knapsack_1_standard_solution)

    def test_knapsack_2_solution(self) -> None:
        self.main_part(func=knapsack_2_solution)

    def test_knapsack_3_solution(self) -> None:
        self.main_part(func=knapsack_3_solution)

    def test_knapsack_4_solution(self) -> None:
        self.main_part(func=knapsack_4_bruteforce_solution)

    def test_knapsack_5_solution(self) -> None:
        self.main_part(func=knapsack_5_dynamic_solution)

    def test_knapsack_6_solution(self) -> None:
        self.main_part(func=knapsack_6_recursive_dynamic_solution)

    def test_knapsack_greedy_solution(self) -> None:
        print('#' * 5, knapsack_greedy_solution.__name__, '#' * 5)
        for data in self.dataset.values():
            start = time.monotonic()
            raw_result = list(knapsack_greedy_solution(*data['input']))
            end = time.monotonic() - start

            self.assertTrue(data['input'].weight_limit >= sum(i.weight for i in raw_result))

            result = sum(i.value for i in raw_result)
            expected_result = sum(i.value for i in data['output']) * 0.7
            self.assertTrue(result >= expected_result)

            print('Knapsack with {} items was completed in {:2f} seconds'.format(len(data['input'].items), end))


if __name__ == '__main__':
    unittest.main()



"""
Examples of python realization of solution for unpacking lists with indefinite depth
"""

from typing import Iterator, Iterable, List

from iteration_utilities import deepflatten
from more_itertools import collapse


# @profile
def outer_flatten_1(array: Iterable) -> List:
    """
    Based on C realization of this solution
    More on:

    https://iteration-utilities.readthedocs.io/en/latest/generated/deepflatten.html

    https://github.com/MSeifert04/iteration_utilities/blob/384948b4e82e41de47fa79fb73efc56c08549b01/src/deepflatten.c
    """
    return deepflatten(array)


# @profile
def outer_flatten_2(array: Iterable) -> List:
    """
    recursive algorithm, vaguely reminiscent of recursion_flatten. Based on next pattern:

    .. code:: python

        try:
            tree = iter(node)
        except TypeError:
            yield node

    more on:
    https://more-itertools.readthedocs.io/en/stable/api.html#more_itertools.collapse
    """
    return collapse(array)


# @profile
def niccolum_flatten(array: Iterable) -> List:
    """
    Non recursive algorithm
    Based on pop/insert elements in current list
    """

    new_array = array[:]
    ind = 0
    while True:
        try:
            while isinstance(new_array[ind], list):
                item = new_array.pop(ind)
                for inner_item in reversed(item):
                    new_array.insert(ind, inner_item)
            ind += 1
        except IndexError:
            break
    return new_array


# @profile
def tishka_flatten(data: Iterable) -> List:
    """
    Non recursive algorithm
    Based on append/extend elements to new list

    """
    nested = True
    while nested:
        new = []
        nested = False
        for i in data:
            if isinstance(i, list):
                new.extend(i)
                nested = True
            else:
                new.append(i)
        data = new
    return data


# @profile
def zart_flatten(a: Iterable) -> List:
    """
    Non recursive algorithm
    Based on pop from old and append elements to new list
    """
    queue, out = [a], []
    while queue:
        elem = queue.pop(-1)
        if isinstance(elem, list):
            queue.extend(elem)
        else:
            out.append(elem)
    return out[::-1]


# @profile
def recursive_flatten_generator(array: Iterable) -> List:
    """
    Recursive algorithm
    Looks like recursive_flatten_iterator, but with extend/append

    """
    lst = []
    for i in array:
        if isinstance(i, list):
            lst.extend(recursive_flatten_generator(i))
        else:
            lst.append(i)
    return lst


# @profile
def recursive_flatten_iterator(arr: Iterable) -> Iterator:
    """
    Recursive algorithm based on iterator
    Usual solution to this problem

    """

    for i in arr:
        if isinstance(i, list):
            yield from recursive_flatten_iterator(i)
        else:
            yield i


# @profile
def tishka_flatten_with_stack(seq: Iterable) -> List:
    """
    Non recursive algorithm
    Based on zart_flatten, but build on try/except pattern
    """
    stack = [iter(seq)]
    new = []
    while stack:
        i = stack.pop()
        try:
            while True:
                data = next(i)
                if isinstance(data, list):
                    stack.append(i)
                    i = iter(data)
                else:
                    new.append(data)
        except StopIteration:
            pass
    return new


if __name__ == '__main__':

    import time

    from data import generate_data, create_data_increasing_depth, create_data_decreasing_depth


    def profile():
        data = generate_data()[-1][1]

        funcs_generated_data = [
            create_data_increasing_depth,
            create_data_decreasing_depth
        ]

        funcs = [
            outer_flatten_1,
            outer_flatten_2,
            niccolum_flatten,
            tishka_flatten,
            zart_flatten,
            recursive_flatten_generator,
            recursive_flatten_iterator,
            tishka_flatten_with_stack
        ]

        for func_generated_data in funcs_generated_data:
            creating_data = func_generated_data(**data)
            for func in funcs:
                list(func(creating_data))
                time.sleep(0.3)

    profile()
